{
   @author(Patrick Michael Kolla-ten Venne [pk] <patrick.kolla@safer-networking.org>)
   @abstract(Provided Pascal class to add MacOS toolbar.)

   @see https://www.lazarusforum.de/viewtopic.php?f=23&t=15347
   @see https://developer.apple.com/documentation/appkit/touch_bar/integrating_a_toolbar_and_touch_bar_into_your_app?language=objc

   Thanks to user Rhyt on the German Lazarus forums for coming up with the idea,
   and providing examples. Without these, I wouldn't have tried or would have
   spent way more time on this.

   @preformatted(
// *****************************************************************************
// Copyright: © 2024 Patrick Michael Kolla-ten Venne. All rights reserved.
// License: BSD 3-Clause Revised License
// Uses code from Rhyt @ https://www.lazarusforum.de/viewtopic.php?f=23&t=15347
// *****************************************************************************
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//      * Redistributions of source code must retain the above copyright
//        notice, this list of conditions and the following disclaimer.
//      * Redistributions in binary form must reproduce the above copyright
//        notice, this list of conditions and the following disclaimer in the
//        documentation and/or other materials provided with the distribution.
//      * Neither the name of the <organization> nor the
//        names of its contributors may be used to endorse or promote products
//        derived from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY Patrick Kolla-ten Venne ``AS IS'' AND ANY
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL Patrick Kolla-ten Venne BE LIABLE FOR ANY
//  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
//  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// *****************************************************************************
// Usage example:
// ---------------------------------------
// Set the custom option -WM11.0
//

//procedure TForm1.addToolbar;
//var
//  toolbar: TMacToolbar;
//  item: TMacToolbarItem;
//begin
//  toolbar := TMacToolbar.Create(Self);
//  item := toolbar.Items.AddItem('refreshToolbarItem', 'Refresh',
//    'Refresh', Self.GetPhoneData);
//  item.ImageDefaulType := NSImageNameRefreshTemplate;
//
//  item := toolbar.Items.AddItem('saveAllChangesItem', 'Save', 'Save',
//    Self.SaveAllChanges);
//  item.ImageDefaulType := NSImageNameActionTemplate;
//
//  item := toolbar.Items.AddItem('AddNew', 'Add', 'Add', Self.MyAddRecord);
//  item.ImageDefaulType := NSImageNameAddTemplate;
//
//  item := toolbar.Items.AddItem('deleteSelectedFrame', 'Delete',
//    'Delete', Self.DeleteSelectedFrame);
//  item.ImageDefaulType := NSImageNameRemoveTemplate;
//
//
//
//  toolbar.Attach;
//end;



// procedure TFormNetworkLeprechaun.AddToolbar;
// var
//    toolbar: TMacToolbar;
//    item: TMacToolbarItem;
// begin
//   toolbar := TMacToolbar.Create(Self);
//   item := toolbar.Items.AddItem('refreshToolbarItem', 'Refresh', 'Refresh', Self.FActionRefresh);
//   item.ImageFilename := 'Resources/rotate.png';
//   item := toolbar.Items.AddItem('connectToolbarItem', 'Connect', 'Connect', Self.FActionConnect);
//   item.ImageFilename := 'Resources/link.png';
//   item := toolbar.Items.AddItem('disconnectToolbarItem', 'Disconnect', 'Disconnect', Self.FActionDisconnect);
//   item.ImageFilename := 'Resources/link-slash.png';
//   toolbar.Attach;
// end;
// *****************************************************************************
// Changelog (new entries first):
// ---------------------------------------
// 2024-02-02  pk   5m  Added the NSHelpers unit instead of copying code from it.
// 2024-02-02  pk  30m  Working code moved from app into separate unit.
// 2024-02-01  pk  30m  Draft.
// *****************************************************************************
// TODO
// - [x] TAction links
// - [x] Editor, for future component
// - [x] Support for TImageList as image source
// - [ ] Finish implemenation of IsCentered
// - [ ] Standard component
// - [ ] Dropdown menus
// - [ ] Delete when active (need subitem support?)
// - [ ] Add when active
// - [ ] ImageList bitmap transparency
// - [ ] For a system image picker, check NSImage to bitmap, https://forum.lazarus.freepascal.org/index.php?topic=32356.0
// *****************************************************************************
   )
}
unit Firefly.MacOS.Toolbar;

{$mode Delphi}{$H+}
{$modeswitch objectivec1}

interface

uses
   Classes,
   LCLClasses,
   SysUtils,
   Forms,
   ActnList,
   Dialogs,
   Menus,
   Graphics,
   ImgList,
   Controls,
   Generics.Collections,
   // CocoaWSForms,
   CocoaAll,
   Firefly.MacOS.ClassHelpers;

type
   TMacToolbarItemImageType = (mtiitNone, mtiitDefault, mtiitFile, mtiitImageList);
   TMacToolbarDisplayMode = (mtdmDefault, mtdmIconAndLabel, mtdmIconOnly, mdtmLabelOnly);
   TMacToolbarSizeMode = (mtsmDefault, mtsmRegular, mtsmSmall);
   TMacToolbarItemType = (mtitItem, mtitItemGroup);

   TMacSegmentStyle = ( // .
      mssAutomatic = 0, // .
      mssRounded = 1, // .
      mssTexturedRounded = 2, // macOS 10.5 +
      mssRoundRect = 3, // .
      mssTexturedSquare = 4, // .
      mssCapsule = 5, // macOS 10.5 +
      mssSmallSquare = 6, // .
      mssSeparated = 8 // macOS 10.10 +
      );

   TMacToolbarDelegate = objcclass;

   TMacToolbarItem = class;
   TMacToolbarItemList = class;

   { TMacToolbarItemActionLink }

   TMacToolbarItemActionLink = class(TActionLink)
   protected
      FClient: TMacToolbarItem;
      procedure AssignClient(AClient: TObject); override;
   protected
      procedure SetCaption(const Value: string); override;
      procedure SetEnabled(Value: Boolean); override;
      procedure SetHelpContext(Value: THelpContext); override;
      procedure SetHint(const Value: string); override;
      procedure SetImageIndex(Value: Integer); override;
      procedure SetVisible(Value: Boolean); override;
      procedure SetOnExecute(Value: TNotifyEvent); override;
   public
      function IsCaptionLinked: Boolean; override;
      function IsEnabledLinked: Boolean; override;
      function IsHelpContextLinked: Boolean; override;
      function IsHintLinked: Boolean; override;
      function IsImageIndexLinked: Boolean; override;
      function IsVisibleLinked: Boolean; override;
   end;

   TMacToolbarItemActionLinkClass = class of TMacToolbarItemActionLink;

   { TMacToolbarItem }

   TMacToolbarItem = class(TLCLComponent)
   private
      FImageIndex: integer;
      FList: TMacToolbarItemList;
      FActionLink: TMacToolbarItemActionLink;
      FCaption: string;
      FDropDownMenu: TPopupmenu;
      FEnabled: boolean;
      FHint: string;
      FImageDefaulType: NSString;
      FImageFilename: string;
      FImageType: TMacToolbarItemImageType;
      FIdentifier: string;
      FIsCentered: boolean;
      FIsDefault: boolean;
      FIsImmovable: boolean;
      FIsSelectable: boolean;
      FItemType: TMacToolbarItemType;
      FSubItems: TMacToolbarItemList;
      function GetAction: TBasicAction;
      procedure SetAction(AValue: TBasicAction);
      procedure SetImageDefaulType(AValue: NSString);
      procedure SetImageFilename(AValue: string);
      procedure DoActionChange(Sender: TObject);
      procedure SetImageIndex(AValue: integer);
   protected
      FMacToolbarItem: NSToolbarItem;
      procedure AssignTo(Dest: TPersistent); override;
      function GetItem(ADelegate: TMacToolbarDelegate): NSToolbarItem;
      function GetActionLinkClass: TMacToolbarItemActionLinkClass; virtual;
      procedure ActionChange(Sender: TObject; CheckDefaults: Boolean); virtual;
      property ActionLink: TMacToolbarItemActionLink read FActionLink write FActionLink;
   public
      constructor Create(TheOwner: TMacToolbarItemList); virtual;
      destructor Destroy; override;
      property ImageType: TMacToolbarItemImageType read FImageType write FImageType;
      property ImageDefaulType: NSString read FImageDefaulType write SetImageDefaulType;
      property ItemType: TMacToolbarItemType read FItemType write FItemType;
      property Owner: TMacToolbarItemList read FList;
   published
      property Identifier: string read FIdentifier write FIdentifier;
      property Caption: string read FCaption write FCaption;
      property Hint: string read FHint write FHint;
      property Action: TBasicAction read GetAction write SetAction;
      property ImageFilename: string read FImageFilename write SetImageFilename;
      property Enabled: boolean read FEnabled write FEnabled;
      property DropDownMenu: TPopupmenu read FDropDownMenu write FDropDownMenu;
      property ImageIndex: integer read FImageIndex write SetImageIndex;
      property IsDefault: boolean read FIsDefault write FIsDefault;
      property IsImmovable: boolean read FIsImmovable write FIsImmovable;
      property IsSelectable: boolean read FIsSelectable write FIsSelectable;
      property IsCentered: boolean read FIsCentered write FIsCentered;
      property SubItems: TMacToolbarItemList read FSubItems;
   end;

   { TMacToolbarItemGroup }

   TMacToolbarItemGroup = class(TMacToolbarItem)
   private
      FSegmentStyle: TMacSegmentStyle;
   protected
      FSegmentedControl: NSSegmentedControl;
      procedure AssignTo(Dest: TPersistent); override;
      function GetItemGroup(ADelegate: TMacToolbarDelegate): NSToolbarItemGroup;
   public
      constructor Create(TheOwner: TMacToolbarItemList); override;
      property SegmentStyle: TMacSegmentStyle read FSegmentStyle write FSegmentStyle;
   end;

   { TMacToolbarItemList }

   TMacToolbarItemList = class(TObjectList<TMacToolbarItem>)
   private
      FImageList: TCustomImageList;
      function GetNSArrayFromStringList(AList: TStringList): NSArray;
      procedure SetImageList(AValue: TCustomImageList);
   protected
      FMacToolbar: NSToolbar;
      procedure GetAllowedItems(AList: TStringList);
      function GetAllowedItemsArray: NSArray;
      procedure GetDefaultItems(AList: TStringList);
      function GetDefaultItemsArray: NSArray;
      procedure GetImmovableItems(AList: TStringList);
      function GetImmovableItemsArray: NSArray;
      function GetSelectableItemsArray: NSArray;
      function GetCenteredItemsArray: NSArray;
      function DoRemove(AIndex: SizeInt; ACollectionNotification: TCollectionNotification): TMacToolbarItem; override;
   public
      function FindItemWithIdentifier(AnIdentifier: NSString; ARecursive: boolean = false): TMacToolbarItem;
      function AddItem: TMacToolbarItem; overload;
      function AddItem(AnIdentifier: string; AnAction: TAction): TMacToolbarItem; overload;
      function AddItem(AnIdentifier, ACaption, AHint: string; AnAction: TAction): TMacToolbarItem; overload;
      function AddItemAfter(ABefore: TMacToolbarItem): TMacToolbarItem;
      function AddItemGroup: TMacToolbarItemGroup; overload;
      function AddItemGroup(AnIdentifier: string): TMacToolbarItemGroup; overload;
      procedure Assign(Source: TMacToolbarItemList);
      property ImageList: TCustomImageList read FImageList write SetImageList;
   end;

   { TMacToolbar }

   TMacToolbar = class(TComponent)
   private
      FAllowsUserCustomization: boolean;
      FAutosavesConfiguration: boolean;
      FImageList: TCustomImageList;
      FShowsBaseLineSeparator: boolean;
      FDisplayMode: TMacToolbarDisplayMode;
      FForm: TCustomForm;
      FItems: TMacToolbarItemList;
      FMacWindow: NSWindow;
      FDelegate: TMacToolbarDelegate;
      FSizeMode: TMacToolbarSizeMode;
      FToolbarIdentifier: string;
      function GetSelectedItem: TMacToolbarItem;
      procedure SetImageList(AValue: TCustomImageList);
      procedure SetSelectedItem(AValue: TMacToolbarItem);
   protected
      FMacToolbar: NSToolbar;
   public
      constructor Create(AOwner: TComponent); override;
      destructor Destroy; override;
      function Attach: boolean; // call once all entries have been added!
      property Items: TMacToolbarItemList read FItems;
      property AllowsUserCustomization: boolean read FAllowsUserCustomization write FAllowsUserCustomization;
      property AutosavesConfiguration: boolean read FAutosavesConfiguration write FAutosavesConfiguration;
      property DisplayMode: TMacToolbarDisplayMode read FDisplayMode write FDisplayMode;
      property SizeMode: TMacToolbarSizeMode read FSizeMode write FSizeMode;
      property ShowsBaseLineSeparator: boolean read FShowsBaseLineSeparator write FShowsBaseLineSeparator;
      property ToolbarIdentifier: string read FToolbarIdentifier write FToolbarIdentifier;
      property SelectedItem: TMacToolbarItem read GetSelectedItem write SetSelectedItem;
      property ImageList: TCustomImageList read FImageList write SetImageList;
   end;

   { TMacToolbarDelegate }

   TMacToolbarDelegate = objcclass(NSObject, NSToolbarDelegateProtocol)
     function toolbar_itemForItemIdentifier_willBeInsertedIntoToolbar({%H-}toolbar: NSToolbar; itemIdentifier: NSString; {%H-}flag: boolean): NSToolbarItem; message 'toolbar:itemForItemIdentifier:willBeInsertedIntoToolbar:';
     function toolbarAllowedItemIdentifiers({%H-}toolbar: NSToolbar): NSArray; message 'toolbarAllowedItemIdentifiers:';
     function toolbarDefaultItemIdentifiers({%H-}toolbar: NSToolbar): NSArray; message 'toolbarDefaultItemIdentifiers:';
     function toolbarImmovableItemIdentifiers({%H-}toolbar: NSToolbar): NSArray; message 'toolbarImmovableItemIdentifiers:';
     function toolbarSelectableItemIdentifiers({%H-}toolbar: NSToolbar): NSArray; message 'toolbarSelectableItemIdentifiers:';
   private
      FItems: TMacToolbarItemList;
      procedure toolbarItemClick(item: NSToolbarItem); message 'toolbarItemClick:';
      procedure toolbarItemGroupClick({%H-}item: NSToolbarItemGroup); message 'toolbarItemGroupClick:';
      procedure segmentedControlClick(control: NSSegmentedControl); message 'segmentedControlClick:';
   public
      function validateToolbarItem(theItem: NSToolbarItem): ObjCBOOL; message 'validateToolbarItem:'; override;
   end;

implementation

{ TMacToolbarItem }

procedure TMacToolbarItem.SetImageFilename(AValue: string);
begin
   FImageFilename := AValue;
   FImageType := mtiitFile;
end;

procedure TMacToolbarItem.DoActionChange(Sender: TObject);
begin
   if Sender = Action then begin
      ActionChange(Sender, False);
   end;
end;

procedure TMacToolbarItem.SetImageIndex(AValue: integer);
begin
   FImageIndex := AValue;
   FImageType := mtiitImageList;
end;

procedure TMacToolbarItem.AssignTo(Dest: TPersistent);
var
   item: TMacToolbarItem;
begin
   if Dest is TMacToolbarItem then begin
      item := TMacToolbarItem(Dest);
      // skip FList!
      item.FActionLink := Self.FActionLink;
      item.FCaption := Self.FCaption;
      item.FImageIndex := Self.FImageIndex;
      item.FDropDownMenu := Self.FDropDownMenu;
      item.FEnabled := Self.FEnabled;
      item.FHint := Self.FHint;
      item.FImageDefaulType := Self.FImageDefaulType;
      item.FImageFilename := Self.FImageFilename;
      item.FImageType := Self.FImageType;
      item.FImageIndex := Self.FImageIndex;
      item.FIdentifier := Self.FIdentifier;
      item.FIsCentered := Self.FIsCentered;
      item.FIsDefault := Self.FIsDefault;
      item.FIsImmovable := Self.FIsImmovable;
      item.FIsSelectable := Self.FIsSelectable;
      item.FItemType := Self.FItemType;
      item.FSubItems.Assign(Self.FSubItems);
   end;
end;

function TMacToolbarItem.GetItem(ADelegate: TMacToolbarDelegate): NSToolbarItem;
var
   img: NSImage;
   ms: TMemoryStream;
   d: NSData;
   bmp: TBitmap;
begin
   Result := NSToolbarItem.alloc.initWithItemIdentifier(NSSTR(Self.Identifier));
   Result.setLabel(NSSTR(Self.Caption));
   Result.setToolTip(NSSTR(Self.Hint));
   Result.setTarget(ADelegate);
   Result.setAction(sel_getUid('toolbarItemClick:'));
   if Assigned(FDropDownMenu) then begin
      // @see https://developer.apple.com/forums/thread/656153
      // Result.setItemMenu(FDropDownMenu.Handle);
   end;
   case Self.ImageType of
      mtiitDefault: begin
         Result.setImage(NSImage.imageNamed(Self.ImageDefaulType));
      end;
      mtiitFile: begin
         img := NSImage.alloc;
         img.initWithContentsOfFile(NSSTR(ExtractFilePath(Application.ExeName) + '../' + Self.ImageFilename));
         Result.setImage(img);
      end;
      mtiitImageList: begin
         if Assigned(Self.FList.FImageList) and (Self.ImageIndex > -1) then begin
            bmp := TBitmap.Create;
            try
               Self.FList.FImageList.GetBitmap(Self.ImageIndex, bmp);
               Result.setImage(bmp.ToNSImage);
            finally
               bmp.Free;
            end;
         end;
      end;
      else begin
         // no image to be set
      end;
   end;
   Self.FMacToolbarItem := Result;
end;

function TMacToolbarItem.GetActionLinkClass: TMacToolbarItemActionLinkClass;
begin
   Result := TMacToolbarItemActionLink;
end;

procedure TMacToolbarItem.ActionChange(Sender: TObject; CheckDefaults: Boolean);
var
  NewAction: TCustomAction;
begin
   if Sender is TCustomAction then begin
      NewAction := TCustomAction(Sender);
      //if (not CheckDefaults) or (AutoCheck = False) then begin
      //   AutoCheck := NewAction.AutoCheck;
      //end;
      if (not CheckDefaults) or (Caption = '') then begin
         Caption := NewAction.Caption;
      end;
      //if (not CheckDefaults) or (Checked = False) then begin
      //   Checked := NewAction.Checked;
      //end;
      if (not CheckDefaults) or (Enabled = True) then begin
         Enabled := NewAction.Enabled;
      end;
      //if (not CheckDefaults) or (HelpContext = 0) then begin
      //   HelpContext := NewAction.HelpContext;
      //end;
      if (not CheckDefaults) or (Hint = '') then begin
         Hint := NewAction.Hint;
      end;
      //if RadioItem and (not CheckDefaults or (GroupIndex = 0)) then begin
      //   GroupIndex := NewAction.GroupIndex;
      //end;
      if (not CheckDefaults) or (ImageIndex = -1) then begin
         ImageIndex := NewAction.ImageIndex;
      end;
      //if (not CheckDefaults) or (ShortCut = scNone) then begin
      //   ShortCut := NewAction.ShortCut;
      //end;
      //if (not CheckDefaults) or (Visible = True) then begin
      //   Visible := NewAction.Visible;
      //end;
  end;
end;

procedure TMacToolbarItem.SetImageDefaulType(AValue: NSString);
begin
   FImageDefaulType := AValue;
   FImageType := mtiitDefault;
end;

function TMacToolbarItem.GetAction: TBasicAction;
begin
   if Assigned(FActionLink) then begin
      Result := FActionLink.Action;
   end else begin
      Result := nil;
   end;
end;

procedure TMacToolbarItem.SetAction(AValue: TBasicAction);
begin
   if AValue = nil then begin
      FActionLink.Free;
      FActionLink := nil;
   end else begin
      if FActionLink = nil then begin
         FActionLink := GetActionLinkClass.Create(Self);
      end;
      FActionLink.Action := AValue;
      FActionLink.OnChange := DoActionChange;
      ActionChange(AValue, csLoading in AValue.ComponentState);
      AValue.FreeNotification(Self);
   end;
end;

constructor TMacToolbarItem.Create(TheOwner: TMacToolbarItemList);
begin
   inherited Create(nil);
   FList := TheOwner;
   FDropDownMenu := nil;
   FImageType := mtiitNone;
   FIsDefault := true;
   FImageIndex := -1;
   FIsImmovable := false;
   FIsSelectable := false;
   FIsCentered := false;
   FEnabled := true;
   FSubItems := TMacToolbarItemList.Create;
   FItemType := mtitItem;
end;

destructor TMacToolbarItem.Destroy;
begin
   FSubItems.Free;
   inherited Destroy;
end;

{ TMacToolbarItemActionLink }

procedure TMacToolbarItemActionLink.AssignClient(AClient: TObject);
begin
   FClient := AClient as TMacToolbarItem;
end;

procedure TMacToolbarItemActionLink.SetCaption(const Value: string);
begin
   if IsCaptionLinked then begin
      FClient.Caption := Value;
   end;
end;

procedure TMacToolbarItemActionLink.SetEnabled(Value: Boolean);
begin
   if IsEnabledLinked then begin
      FClient.Enabled := Value;
   end;
end;

procedure TMacToolbarItemActionLink.SetHelpContext(Value: THelpContext);
begin
   if IsHelpContextLinked then begin
      // FClient.HelpContext := Value;
   end;
end;

procedure TMacToolbarItemActionLink.SetHint(const Value: string);
begin
   if IsHintLinked then begin
      FClient.Hint := Value;
   end;
end;

procedure TMacToolbarItemActionLink.SetImageIndex(Value: Integer);
begin
   if IsImageIndexLinked then begin
      FClient.ImageIndex := Value;
   end;
end;

procedure TMacToolbarItemActionLink.SetVisible(Value: Boolean);
begin
   if IsVisibleLinked then begin
      //FClient.Visible := Value;
   end;
end;

procedure TMacToolbarItemActionLink.SetOnExecute(Value: TNotifyEvent);
begin
   inherited;
end;

function TMacToolbarItemActionLink.IsCaptionLinked: Boolean;
begin
   Result := inherited IsCaptionLinked
     and (AnsiCompareText(FClient.Caption, (Action as TCustomAction).Caption)=0);
end;

function TMacToolbarItemActionLink.IsEnabledLinked: Boolean;
begin
   Result := inherited IsEnabledLinked
     and (FClient.Enabled = (Action as TCustomAction).Enabled);
end;

function TMacToolbarItemActionLink.IsHelpContextLinked: Boolean;
begin
   Result := inherited IsHelpContextLinked;
//             and (FClient.HelpContext = (Action as TCustomAction).HelpContext);
end;

function TMacToolbarItemActionLink.IsHintLinked: Boolean;
begin
   Result := inherited IsHintLinked
             and (FClient.Hint = (Action as TCustomAction).Hint);
end;

function TMacToolbarItemActionLink.IsImageIndexLinked: Boolean;
begin
   Result := inherited IsImageIndexLinked
             and (FClient.ImageIndex = (Action as TCustomAction).ImageIndex);
end;

function TMacToolbarItemActionLink.IsVisibleLinked: Boolean;
begin
   Result := inherited IsVisibleLinked;
//             and (FClient.Visible = (Action as TCustomAction).Visible);
end;

{ TMacToolbarItemGroup }

procedure TMacToolbarItemGroup.AssignTo(Dest: TPersistent);
var
   item: TMacToolbarItemGroup;
begin
   inherited AssignTo(Dest);
   if Dest is TMacToolbarItemGroup then begin
      item := TMacToolbarItemGroup(Dest);
      item.FSegmentedControl := Self.FSegmentedControl;
      item.FSegmentStyle := Self.FSegmentStyle;
   end;
end;

function TMacToolbarItemGroup.GetItemGroup(ADelegate: TMacToolbarDelegate): NSToolbarItemGroup;
var
   a: array of NSToolbarItem = [];
   i: integer = 0;
   item: TMacToolbarItem;
   nsa: NSArray;
begin
   Result := NSToolbarItemGroup.alloc.initWithItemIdentifier(NSSTR(Self.Identifier));
   Result.setLabel(NSSTR(Self.Caption));
   Result.setTarget(ADelegate);
   Result.setAction(sel_getUid('toolbarItemGroupClick:'));
   SetLength(a, Self.SubItems.Count);
   for item in Self.SubItems do begin
      a[i] := item.GetItem(ADelegate);
      Inc(i);
   end;
   nsa := NSArray.arrayWithObjects_count(idptr(@a[0]), Length(a));
   Result.setSubitems(nsa);

   FSegmentedControl := (NSSegmentedControl.alloc).init;
   FSegmentedControl.setSegmentCount(Length(a));
   for i := 0 to Pred(Length(a)) do begin
      FSegmentedControl.setWidth_forSegment(40, i);
      FSegmentedControl.setImage_forSegment(a[i].image, i);
      // FSegmentedControl.setMenu_forSegment(NSMenu(Self.SubItems[i].DropDownMenu.Handle), i);
      // FSegmentedControl.setLabel_forSegment(a[i].label_, i);
   end;
   // FSegmentedControl.setMenu(NSMenu(Self.SubItems[0].DropDownMenu.Handle));
   FSegmentedControl.setSegmentStyle(NSSegmentStyle(FSegmentStyle));
   FSegmentedControl.setTarget(ADelegate);
   FSegmentedControl.setAction(sel_getUid('segmentedControlClick:'));
   Result.setView(FSegmentedControl);
end;

constructor TMacToolbarItemGroup.Create(TheOwner: TMacToolbarItemList);
begin
   inherited;
   FSegmentStyle := mssAutomatic;
end;

{ TMacToolbarItemList }

function TMacToolbarItemList.FindItemWithIdentifier(AnIdentifier: NSString;
   ARecursive: boolean): TMacToolbarItem;
var
   i: TMacToolbarItem;
begin
   Result := nil;
   for i in Self do begin
      if 0 = AnIdentifier.compare(NSSTR(i.Identifier)) then begin
         Result := i;
         Exit;
      end;
      if ARecursive then begin
         Result := i.SubItems.FindItemWithIdentifier(AnIdentifier, true);
         if Assigned(Result) then begin
            Exit;
         end;
      end;
   end;
end;

function TMacToolbarItemList.AddItem(AnIdentifier, ACaption, AHint: string; AnAction: TAction): TMacToolbarItem;
begin
   Result := AddItem(AnIdentifier, AnAction);
   Result.Caption := ACaption;
   Result.Hint := AHint;
end;

function TMacToolbarItemList.AddItemAfter(ABefore: TMacToolbarItem): TMacToolbarItem;
begin
   Result := TMacToolbarItem.Create(Self);
   Self.Insert(Self.IndexOf(ABefore) + 1, Result);
end;

function TMacToolbarItemList.AddItem(AnIdentifier: string; AnAction: TAction): TMacToolbarItem;
begin
   Result := AddItem;
   Result.Identifier := AnIdentifier;
   Result.Action := AnAction;
end;

function TMacToolbarItemList.AddItem: TMacToolbarItem;
begin
   Result := TMacToolbarItem.Create(Self);
   Result.SubItems.FMacToolbar := FMacToolbar;
   Result.SubItems.ImageList := Self.ImageList;
   Self.Add(Result);
end;

function TMacToolbarItemList.AddItemGroup(AnIdentifier: string): TMacToolbarItemGroup;
begin
   Result := AddItemGroup;
   Result.Identifier := AnIdentifier;
end;

procedure TMacToolbarItemList.Assign(Source: TMacToolbarItemList);
var
   item: TMacToolbarItem;
   new: TMacToolbarItem;
   newGroup: TMacToolbarItemGroup;
begin
   Self.Clear;
   Self.ImageList := Source.ImageList;
   for item in Source do begin
      if item is TMacToolbarItemGroup then begin
         newGroup := Self.AddItemGroup;
         newGroup.Assign(item);
      end else begin
         new := Self.AddItem;
         new.Assign(item);
      end;
   end;
end;

function TMacToolbarItemList.AddItemGroup: TMacToolbarItemGroup;
begin
   Result := TMacToolbarItemGroup.Create(Self);
   Result.SubItems.FMacToolbar := FMacToolbar;
   Result.SubItems.ImageList := Self.ImageList;
   Result.ItemType := mtitItemGroup;
   Self.Add(Result);
end;

function TMacToolbarItemList.GetNSArrayFromStringList(AList: TStringList): NSArray;
var
   a: array of NSString = [];
   i: integer = 0;
   s: string;
begin
   SetLength(a, AList.Count);
   for s in AList do begin
      a[i] := NSSTR(s);
      Inc(i);
   end;
   Result := NSArray.arrayWithObjects_count(idptr(@a[0]), Length(a));
end;

procedure TMacToolbarItemList.SetImageList(AValue: TCustomImageList);
var
   item: TMacToolbarItem;
begin
   FImageList := AValue;
   for item in Self do begin
      item.SubItems.ImageList := AValue;
   end;
end;

procedure TMacToolbarItemList.GetAllowedItems(AList: TStringList);
var
   item: TMacToolbarItem;
   sl: TStringList;
begin
   for item in Self do begin
      AList.Add(item.Identifier);
      if item.SubItems.Count > 0 then begin
         sl := TStringList.Create;
         try
            item.SubItems.GetAllowedItems(sl);
            AList.AddStrings(sl);
         finally
            sl.Free;
         end;
      end;
   end;
end;

{
  Returns an array of default items.
}
function TMacToolbarItemList.GetDefaultItemsArray: NSArray;
var
   sl: TStringList;
begin
    sl := TStringList.Create;
   try
      Self.GetDefaultItems(sl);
      Result := GetNSArrayFromStringList(sl);
   finally
      sl.Free;
   end;
end;

procedure TMacToolbarItemList.GetImmovableItems(AList: TStringList);
var
   item: TMacToolbarItem;
   sl: TStringList;
begin
   for item in Self do begin
      if item.IsImmovable then begin
         AList.Add(item.Identifier);
         if item.SubItems.Count > 0 then begin
            sl := TStringList.Create;
            try
               item.SubItems.GetImmovableItems(sl);
               AList.AddStrings(sl);
            finally
               sl.Free;
            end;
         end;
      end;
   end;
end;

{
  Returns an array of allowed (all) items.
}
function TMacToolbarItemList.GetAllowedItemsArray: NSArray;
var
   sl: TStringList;
begin
   sl := TStringList.Create;
   try
      Self.GetAllowedItems(sl);
      Result := GetNSArrayFromStringList(sl);
   finally
      sl.Free;
   end;
end;

procedure TMacToolbarItemList.GetDefaultItems(AList: TStringList);
var
   item: TMacToolbarItem;
   sl: TStringList;
begin
   for item in Self do begin
      if item.IsDefault then begin
         AList.Add(item.Identifier);
         if item.SubItems.Count > 0 then begin
            sl := TStringList.Create;
            try
               item.SubItems.GetDefaultItems(sl);
               AList.AddStrings(sl);
            finally
               sl.Free;
            end;
         end;
      end;
   end;
end;

{
  Returns an array of immovable items.
}
function TMacToolbarItemList.GetImmovableItemsArray: NSArray;

var
   sl: TStringList;
begin
    sl := TStringList.Create;
   try
      Self.GetImmovableItems(sl);
      Result := GetNSArrayFromStringList(sl);
   finally
      sl.Free;
   end;
end;

{
  Returns an array of selectable items.
}
function TMacToolbarItemList.GetSelectableItemsArray: NSArray;
var
   item: TMacToolbarItem;
   a: array of NSString = [];
   i: integer = 0;
begin
   for item in Self do begin
      if item.IsSelectable then begin
         SetLength(a, Succ(i));
         a[i] := NSSTR(item.Identifier);
         Inc(i);
      end;
   end;
   Result := NSArray.arrayWithObjects_count(idptr(@a[0]), Length(a));
end;

{
  Returns an array of items defined to be centered.

  This is compatible with macOS 13.0+

  The property centeredItemIdentifiers is Not defined in FreePascal yet.

  @see https://developer.apple.com/documentation/appkit/nstoolbar/3969238-centereditemidentifiers?language=objc
}
function TMacToolbarItemList.GetCenteredItemsArray: NSArray;
var
   item: TMacToolbarItem;
   a: array of NSString = [];
   i: integer = 0;
begin
   for item in Self do begin
      if item.IsCentered then begin
         SetLength(a, Succ(i));
         a[i] := NSSTR(item.Identifier);
         Inc(i);
      end;
   end;
   Result := NSArray.arrayWithObjects_count(idptr(@a[0]), Length(a));
end;

function TMacToolbarItemList.DoRemove(AIndex: SizeInt; ACollectionNotification: TCollectionNotification): TMacToolbarItem;
begin
   Result := inherited DoRemove(AIndex, ACollectionNotification);
   // TODO : if toolbar is live, remove from toolbar
   FMacToolbar.removeItemAtIndex(AIndex);
end;

{ TMacToolbar }

function TMacToolbar.GetSelectedItem: TMacToolbarItem;
begin
   Result := FItems.FindItemWithIdentifier(FMacToolbar.selectedItemIdentifier, true);
end;

procedure TMacToolbar.SetImageList(AValue: TCustomImageList);
begin
   FImageList := AValue;
   FItems.ImageList := AValue;
end;

procedure TMacToolbar.SetSelectedItem(AValue: TMacToolbarItem);
begin
   FMacToolbar.setSelectedItemIdentifier(NSSTR(AValue.FIdentifier));
end;

constructor TMacToolbar.Create(AOwner: TComponent);
begin
   inherited;
   if AOwner is TCustomForm then begin
      FForm := TCustomForm(AOwner);
   end else begin
      FForm := nil;
   end;
   FItems := TMacToolbarItemList.Create;
   FDisplayMode := mtdmDefault;
   FSizeMode := mtsmDefault;
   FAllowsUserCustomization := true;
   FShowsBaseLineSeparator := false; // default
   FToolbarIdentifier := AOwner.ClassName + 'LazarusMacToolbar' + FormatDateTime('yyyymmddhhnnsszzz', Now);
end;

destructor TMacToolbar.Destroy;
begin
   FItems.Free;
   inherited Destroy;
end;

function TMacToolbar.Attach: boolean;
begin
   Result := false;
   if not Assigned(FForm) then begin
      Exit;
   end;
   try
      FMacWindow := {%H-}NSView(FForm.Handle).window;
      FMacToolbar := NSToolbar.alloc;
      {$IF MAC_OS_X_VERSION_MIN_REQUIRED >= 1100}
      FMacToolbar.initWithIdentifier(NSSTR(FToolbarIdentifier));
      {$ELSE}
      FMacToolbar.init;
      {$ENDIF}
      FDelegate := TMacToolbarDelegate.alloc.init;
      FDelegate.FItems := Self.Items;
      FMacToolbar.setDelegate(FDelegate);
      FMacToolbar.setAllowsUserCustomization(FAllowsUserCustomization);
      FMacToolbar.setAutosavesConfiguration(FAutosavesConfiguration);
      case FDisplayMode of
         mtdmDefault: begin
            FMacToolbar.setDisplayMode(NSToolbarDisplayModeDefault);
         end;
         mtdmIconAndLabel: begin
            FMacToolbar.setDisplayMode(NSToolbarDisplayModeIconAndLabel);
         end;
         mtdmIconOnly: begin
            FMacToolbar.setDisplayMode(NSToolbarDisplayModeIconOnly);
         end;
         mdtmLabelOnly: begin
            FMacToolbar.setDisplayMode(NSToolbarDisplayModeLabelOnly);
         end;
      end;
      case FSizeMode of
         mtsmDefault: begin
            FMacToolbar.setSizeMode(NSToolbarSizeModeDefault);
         end;
         mtsmRegular: begin
            FMacToolbar.setSizeMode(NSToolbarSizeModeRegular);
         end;
         mtsmSmall: begin
            FMacToolbar.setSizeMode(NSToolbarSizeModeSmall);
         end;
      end;
      FMacToolbar.setShowsBaselineSeparator(FShowsBaseLineSeparator);
      FMacWindow.setToolbar(FMacToolbar);
      FItems.FMacToolbar := FMacToolbar;
      Result := true;
   except
      Result := false;
   end;
end;

{ TMacToolbarDelegate }

{
  Asks the delegate for the toolbar item associated with the specified identifier.

  Use this method to create new NSToolbarItem objects when the toolbar asks for
  them. If your toolbar item uses a custom view, make sure that view is fully
  configured before you return the item. The toolbar becomes the owner of the
  returned item, but can display the item either in the toolbar or the
  customization palette.

  Don’t recycle toolbar items; always provide a new instance, even if the
  toolbar previously asked for an item with the same identifier.

  Even though this is an optional method, you must implement it if you create
  the toolbar programatically.

  @see https://developer.apple.com/documentation/appkit/nstoolbardelegate/1516985-toolbar?language=objc

  @param toolbar NSToolbar        The toolbar for which the item is being requested.
  @param idemIdentifier NSString  The identifier for the requested item.
  @param flag boolean             YES if the toolbar will insert the item immediately.
                                  If this parameter is NO, provide a canonical
                                  representation for the item. For example, provide
                                  a version of the item suitable for display in
                                  the toolbar customization sheet.

  @return NSToolbarItem  A new NSToolbarItem object, or nil if no toolbar item is available
                         for the specified identifier.
}
function TMacToolbarDelegate.toolbar_itemForItemIdentifier_willBeInsertedIntoToolbar(toolbar: NSToolbar; itemIdentifier: NSString; flag: boolean): NSToolbarItem;
var
   item: TMacToolbarItem;
begin
   Result := nil;
   item := FItems.FindItemWithIdentifier(itemIdentifier);
   if Assigned(item) then begin
      case item.ItemType of
         mtitItem: begin
            Result := item.GetItem(Self);
         end;
         mtitItemGroup: begin
            if item is TMacToolbarItemgroup then begin
               Result := TMacToolbarItemgroup(item).GetItemGroup(Self);
            end;
         end;
      end;
   end;
      // 1. custom buttons would use toolbarItem.setView
      // can we use this for e.g. TListbox or something similar?
      // 2. possible support of TPopupMenu: use this:
      (*
      let menuItem: NSMenuItem = NSMenuItem()
      menuItem.submenu = nil
      menuItem.title = label
      toolbarItem.menuFormRepresentation = menuItem
      *)
end;


{
  Asks the delegate to provide the items allowed on the toolbar.

  Include all of your toolbar’s items, including standard ones defined by
  NSToolbarIdentifier. The array must include all of the default menu items in
  your toolbar.

  Even though this is an optional method, you must implement it if you create
  the toolbar programatically.

  @see https://developer.apple.com/documentation/appkit/nstoolbardelegate/1516995-toolbaralloweditemidentifiers?language=objc

  @param toolbar NSToolbar  The toolbar whose allowed item identifiers are to be returned.

  @return NSArray<NSToolbarItemIdentifier>  An array of toolbar item identifiers,
          each of which represents an item that appears in the customization palette.
          Arrange the identifiers in the order you want them to appear in the
          palette, with the first item appearing on the palette’s leading edge.
}
function TMacToolbarDelegate.toolbarAllowedItemIdentifiers(toolbar: NSToolbar): NSArray;
begin
   Result := FItems.GetAllowedItemsArray;
end;

{
  Asks the delegate to provide the default items to display on the toolbar.

  The toolbar calls this method when user settings don’t contain any custom
  configuration data for the toolbar. The toolbar also calls it to initialize
  the customization palette’s contents.

  Even though this is an optional method, you must implement it if you create
  the toolbar programatically. If you configure your toolbar in Interface
  Builder, the configuration there provides the default items.

  @see https://developer.apple.com/documentation/appkit/nstoolbardelegate/1516944-toolbardefaultitemidentifiers?language=objc

  @param toolbar NSToolbar  The toolbar whose default item identifiers are to be returned.

  @return NSArray<NSToolbarItemIdentifier>  An array of toolbar item identifiers,
          each of which represents an item that appears in the default toolbar.
          Arrange the identifiers in the order you want them to appear in the
          toolbar, with the first item appearing on the toolbar’s leading edge.
}
function TMacToolbarDelegate.toolbarDefaultItemIdentifiers(toolbar: NSToolbar): NSArray;
begin
   Result := FItems.GetDefaultItemsArray;
end;

{
  Asks the delegate to provide the items that people can’t remove from the
  toolbar or rearrange during the customization process.

  Implement this method in your delegate and return any items you don’t want
  people to remove or rearrange. If you don’t implement this method, the toolbar
  lets people rearrange and remove all toolbar items.

  @see https://developer.apple.com/documentation/appkit/nstoolbardelegate/3969240-toolbarimmovableitemidentifiers?language=objc

  @param toolbar NSToolbar  The toolbar that contains the items.

  @return NSArray<NSToolbarItemIdentifier>  The set of item identifiers that people
          can’t remove from the toolbar or move to other locations in the toolbar.
          Return an empty set to let someone customize all toolbar items.
}

function TMacToolbarDelegate.toolbarImmovableItemIdentifiers(toolbar: NSToolbar): NSArray;
begin
   Result := FItems.GetImmovableItemsArray;
end;


{
  Asks the delegate to provide the set of selectable items in the toolbar.

  An array of item identifiers, each of which corresponds to an NSToolbarItem
  that should display a selection indicator in the specified toolbar.

  Use this method to return the complete list of toolbar items that support
  selection. When someone selects one of the returned items, the toolbar
  automatically displays that item with a visual highlight. The toolbar also
  places the currently selected item in its selectedItemIdentifier property.

  @see https://developer.apple.com/documentation/appkit/nstoolbardelegate/1516981-toolbarselectableitemidentifiers?language=objc

  @param toolbar NSToolbar  The toolbar that contains the items.

  @return NSArray<NSToolbarItemIdentifier>  An array of item identifiers, each
          of which corresponds to an NSToolbarItem that should display a
          selection indicator in the specified toolbar.
}

function TMacToolbarDelegate.toolbarSelectableItemIdentifiers(toolbar: NSToolbar): NSArray;
begin
   Result := FItems.GetSelectableItemsArray;
end;

{
  If this method is implemented and returns NO, NSToolbar will disable theItem; re
  turning YES causes theItem to be enabled.

  If the receiver is the target for the actions of multiple toolbar items,
  it’s necessary to determine which toolbar item theItem refers to by testing
  the itemIdentifier.

  @see https://developer.apple.com/documentation/objectivec/nsobject/1524282-validatetoolbaritem
}
function TMacToolbarDelegate.validateToolbarItem(theItem: NSToolbarItem): ObjCBOOL;
var
   mtitem: TMacToolbarItem;
begin
   Result := false;
   if not Assigned(theItem) then begin
      Exit;
   end;
   mtitem := FItems.FindItemWithIdentifier(theItem.itemIdentifier);
   if Assigned(mtitem) then begin
      Result := mtitem.Enabled;
   end;
end;

procedure TMacToolbarDelegate.toolbarItemClick(item: NSToolbarItem);
var
   mtitem: TMacToolbarItem;
begin
   if not Assigned(item) then begin
      Exit;
   end;
   mtitem := FItems.FindItemWithIdentifier(item.itemIdentifier);
   if Assigned(mtitem) then begin
      if Assigned(mtitem.Action) then begin
         mtitem.Action.Execute;
      end;
   end;
end;

procedure TMacToolbarDelegate.toolbarItemGroupClick(item: NSToolbarItemGroup);
begin
   // seems to not be triggered, use segmentedControlClick instead.
end;

procedure TMacToolbarDelegate.segmentedControlClick(control: NSSegmentedControl);
var
   item: TMacToolbarItem;
   subItem: TMacToolbarItem;
   i: integer;
begin
   for item in FItems do begin
      if (item is TMacToolbarItemGroup) then begin
         if (TMacToolbarItemGroup(item).FSegmentedControl = control) then begin
            i := control.selectedSegment;
            if (i < item.SubItems.Count) then begin
               subItem := item.SubItems[i];
               if Assigned(subItem.Action) then begin
                  subItem.Action.Execute;
               end;
            end;
         end;
      end;
   end;
end;

end.
